@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2010-2012 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2010-2014 -- leonerd@leonerd.org.uk
 
 package IO::Socket::IP;
 
@@ -9,7 +9,7 @@ use strict;
 use warnings;
 use base qw( IO::Socket );
 
-our $VERSION = '0.16';
+our $VERSION = '0.25';
 
 use Carp;
 
@@ -58,8 +58,7 @@ my $IPv6_re = do {
 
 =head1 NAME
 
-C<IO::Socket::IP> - A drop-in replacement for C<IO::Socket::INET> supporting
-both IPv4 and IPv6
+C<IO::Socket::IP> - Family-neutral IP socket supporting both IPv4 and IPv6
 
 =head1 SYNOPSIS
 
@@ -80,7 +79,7 @@ both IPv4 and IPv6
 =head1 DESCRIPTION
 
 This module provides a protocol-independent way to use IPv4 and IPv6 sockets,
-as a drop-in replacement for L<IO::Socket::INET>. Most constructor arguments
+intended as a replacement for L<IO::Socket::INET>. Most constructor arguments
 and methods are provided in a backward-compatible way. For a list of known
 differences, see the C<IO::Socket::INET> INCOMPATIBILITES section below.
 
@@ -124,14 +123,14 @@ sub import
 
    foreach ( @_ ) {
       if( $_ eq "-register" ) {
-         $pkg->register_domain( AF_INET );
-         $pkg->register_domain( $AF_INET6 ) if defined $AF_INET6;
+         IO::Socket::IP::_ForINET->register_domain( AF_INET );
+         IO::Socket::IP::_ForINET6->register_domain( $AF_INET6 ) if defined $AF_INET6;
       }
       else {
          push @symbols, $_;
       }
    }
-   
+
    @_ = ( $pkg, @symbols );
    goto &IO::Socket::import;
 }
@@ -235,10 +234,12 @@ either in string name or numeric form.
 
 =item GetAddrInfoFlags => INT
 
-More flags to pass to the C<getaddrinfo()> function. These flags will be
-combined with C<AI_ADDRCONFIG>, and if the C<Listen> argument is given,
-C<AI_PASSIVE>. For more information see the documentation about
-C<getaddrinfo()> in the L<Socket> module.
+More flags to pass to the C<getaddrinfo()> function. If not supplied, a
+default of C<AI_ADDRCONFIG> will be used.
+
+These flags will be combined with C<AI_PASSIVE> if the C<Listen> argument is
+given. For more information see the documentation about C<getaddrinfo()> in
+the L<Socket> module.
 
 =item Listen => INT
 
@@ -352,9 +353,9 @@ sub configure
       my $host    = $type . 'Host';
       my $service = $type . 'Service';
 
-      if (exists $arg->{$host} && !exists $arg->{$service}) {
-         defined $arg->{$host} or next;
+      if( defined $arg->{$host} ) {
          ( $arg->{$host}, my $s ) = $self->split_addr( $arg->{$host} );
+         # IO::Socket::INET compat - *Host parsed port always takes precedence
          $arg->{$service} = $s if defined $s;
       }
    }
@@ -371,7 +372,12 @@ sub _configure
    my @localinfos;
    my @peerinfos;
 
-   $hints{flags} = ( $arg->{GetAddrInfoFlags} || 0 ) | $AI_ADDRCONFIG;
+   if( defined $arg->{GetAddrInfoFlags} ) {
+      $hints{flags} = $arg->{GetAddrInfoFlags};
+   }
+   else {
+      $hints{flags} = $AI_ADDRCONFIG;
+   }
 
    if( defined( my $family = $arg->{Family} ) ) {
       $hints{family} = $family;
@@ -391,7 +397,7 @@ sub _configure
       $hints{protocol} = $proto;
    }
 
-   # To maintain compatibilty with IO::Socket::INET, imply a default of
+   # To maintain compatibility with IO::Socket::INET, imply a default of
    # SOCK_STREAM + IPPROTO_TCP if neither hint is given
    if( !defined $hints{socktype} and !defined $hints{protocol} ) {
       $hints{socktype} = SOCK_STREAM;
@@ -505,6 +511,28 @@ sub _configure
       }
    }
 
+   if( !@infos ) {
+      # If there was a Family hint then create a plain unbound, unconnected socket
+      # If there wasn't, use getaddrinfo()'s AI_ADDRCONFIG side-effect to guess a
+      # suitable family first.
+      if( !defined $hints{family} ) {
+         my ( $err, $addrinfo ) = getaddrinfo( "", "0", \%hints );
+         if( $err ) {
+            $@ = "$err";
+            $! = EINVAL;
+            return;
+         }
+
+         $hints{family} = $addrinfo->{family};
+      }
+
+      @infos = ( {
+         family   => $hints{family},
+         socktype => $hints{socktype},
+         protocol => $hints{protocol},
+      } );
+   }
+
    # In the nonblocking case, caller will be calling ->setup multiple times.
    # Store configuration in the object for the ->setup method
    # Yes, these are messy. Sorry, I can't help that...
@@ -626,7 +654,7 @@ sub connected
    my $self = shift;
    return defined $self->fileno &&
           !${*$self}{io_socket_ip_connect_in_progress} &&
-          defined $self->peername;
+          defined getpeername( $self ); # ->peername caches, we need to detect disconnection
 }
 
 =head1 METHODS
@@ -779,7 +807,7 @@ sub peeraddr { my $self = shift; _unpack_sockaddr $self->peername }
 sub accept
 {
    my $self = shift;
-   my ( $new, $peer ) = $self->SUPER::accept or return;
+   my ( $new, $peer ) = $self->SUPER::accept( @_ ) or return;
 
    ${*$new}{$_} = ${*$self}{$_} for qw( io_socket_domain io_socket_type io_socket_proto );
 
@@ -800,6 +828,43 @@ sub socket
    dup2( $tmph->fileno, $self->fileno ) or die "Unable to dup2 $tmph onto $self - $!";
 }
 
+# Versions of IO::Socket before 1.35 may leave socktype undef if from, say, an
+#   ->fdopen call. In this case we'll apply a fix
+BEGIN {
+   if( $IO::Socket::VERSION < 1.35 ) {
+      *socktype = sub {
+         my $self = shift;
+         my $type = $self->SUPER::socktype;
+         if( !defined $type ) {
+            $type = $self->sockopt( Socket::SO_TYPE() );
+         }
+         return $type;
+      };
+   }
+}
+
+=head2 $inet = $sock->as_inet
+
+Returns a new L<IO::Socket::INET> instance wrapping the same filehandle. This
+may be useful in cases where it is required, for backward-compatibility, to
+have a real object of C<IO::Socket::INET> type instead of C<IO::Socket::IP>.
+The new object will wrap the same underlying socket filehandle as the
+original, so care should be taken not to continue to use both objects
+concurrently. Ideally the original C<$sock> should be discarded after this
+method is called.
+
+This method checks that the socket domain is C<PF_INET> and will throw an
+exception if it isn't.
+
+=cut
+
+sub as_inet
+{
+   my $self = shift;
+   croak "Cannot downgrade a non-PF_INET socket to IO::Socket::INET" unless $self->sockdomain == AF_INET;
+   return IO::Socket::INET->new_from_fd( $self->fileno, "r+" );
+}
+
 =head1 NON-BLOCKING
 
 If the constructor is passed a defined but false value for the C<Blocking>
@@ -875,8 +940,7 @@ To support the C<IO::Socket::INET> API, the host and port information may be
 passed in a single string rather than as two separate arguments.
 
 If either C<LocalHost> or C<PeerHost> (or their C<...Addr> synonyms) have any
-of the following special forms, and C<LocalService> or C<PeerService> (or
-their C<...Port> synonyms) are absent, special parsing is applied.
+of the following special forms then special parsing is applied.
 
 The value of the C<...Host> argument will be split to give both the hostname
 and port (or service name):
@@ -897,6 +961,10 @@ service name and number, in a form such as
 In this case, the name (C<http>) will be tried first, but if the resolver does
 not understand it then the port number (C<80>) will be used instead.
 
+If the C<...Host> argument is in this special form and the corresponding
+C<...Service> or C<...Port> argument is also defined, the one parsed from
+the C<...Host> argument will take precedence and the other will be ignored.
+
 =head2 ( $host, $port ) = IO::Socket::IP->split_addr( $addr )
 
 Utility method that provides the parsing functionality described above.
@@ -933,6 +1001,62 @@ sub split_addr
    return ( $addr, undef );
 }
 
+=head2 $addr = IO::Socket::IP->join_addr( $host, $port )
+
+Utility method that performs the reverse of C<split_addr>, returning a string
+formed by joining the specified host address and port number. The host address
+will be wrapped in C<[]> brackets if required (because it is a raw IPv6
+numeric address).
+
+This can be especially useful when combined with the C<sockhost_service> or
+C<peerhost_service> methods.
+
+ say "Connected to ", IO::Socket::IP->join_addr( $sock->peerhost_service );
+
+=cut
+
+sub join_addr
+{
+   shift;
+   my ( $host, $port ) = @_;
+
+   $host = "[$host]" if $host =~ m/:/;
+
+   return join ":", $host, $port if defined $port;
+   return $host;
+}
+
+# Since IO::Socket->new( Domain => ... ) will delete the Domain parameter
+# before calling ->configure, we need to keep track of which it was
+
+package # hide from indexer
+   IO::Socket::IP::_ForINET;
+use base qw( IO::Socket::IP );
+
+sub configure
+{
+   # This is evil
+   my $self = shift;
+   my ( $arg ) = @_;
+
+   bless $self, "IO::Socket::IP";
+   $self->configure( { %$arg, Family => Socket::AF_INET() } );
+}
+
+package # hide from indexer
+   IO::Socket::IP::_ForINET6;
+use base qw( IO::Socket::IP );
+
+sub configure
+{
+   # This is evil
+   my $self = shift;
+   my ( $arg ) = @_;
+
+   bless $self, "IO::Socket::IP";
+   $self->configure( { %$arg, Family => Socket::AF_INET6() } );
+}
+
 =head1 C<IO::Socket::INET> INCOMPATIBILITES
 
 =over 4
